package net.apunch.blacksmith;

import net.apunch.blacksmith.util.Settings;
import net.apunch.blacksmith.util.Settings.Setting;
import net.citizensnpcs.api.CitizensAPI;
import net.citizensnpcs.api.util.DataKey;
import net.milkbowl.vault.economy.Economy;
import org.bukkit.Bukkit;
import org.bukkit.ChatColor;
import org.bukkit.Material;
import org.bukkit.NamespacedKey;
import org.bukkit.command.Command;
import org.bukkit.command.CommandSender;
import org.bukkit.enchantments.Enchantment;
import org.bukkit.entity.Player;
import org.bukkit.inventory.ItemStack;
import org.bukkit.inventory.meta.Damageable;
import org.bukkit.plugin.Plugin;
import org.bukkit.plugin.RegisteredServiceProvider;
import org.bukkit.plugin.java.JavaPlugin;
import org.jetbrains.annotations.NotNull;
import regalowl.hyperconomy.HyperAPI;
import regalowl.hyperconomy.HyperConomy;
import regalowl.hyperconomy.bukkit.BukkitConnector;
import regalowl.hyperconomy.inventory.HItemStack;
import regalowl.hyperconomy.tradeobject.TradeObject;

import java.util.ArrayList;
import java.util.Objects;
import java.util.logging.Level;

public class BlacksmithPlugin extends JavaPlugin {
    private Settings config;
    private Economy economy;
    private HyperAPI hyperAPI;
    private BukkitConnector bukCon;
    private boolean useHyperAPI = false;

    @Override
    public void onDisable() {
        //	config.save();

        getLogger().log(Level.INFO, " v" + getDescription().getVersion() + " disabled.");
    }

    @Override
    public void onEnable() {
        config = new Settings(this);
        config.load();
        /* Setup Hyperconomy (Soft-Depend only, so this is completely optional!)    
            Hyperconomy uses your favorite Vault-compatible economy system
            and calculates prices for items based on supply and demand on the fly.
            This is only used to get the cost of a repair.*/
        if (Bukkit.getPluginManager().getPlugin("HyperConomy") != null) {
            getServer().getLogger().log(Level.INFO, "Found HyperConomy! Using that for calculating prices, base-prices and price-per-durability-point in the Blacksmith config.yml will NOT be used!");
            this.useHyperAPI = true;
            Plugin hcPlugin = getServer().getPluginManager().getPlugin("HyperConomy");
            bukCon = (BukkitConnector) hcPlugin;
            HyperConomy hc = Objects.requireNonNull(bukCon).getHC();
            this.hyperAPI = (HyperAPI) hc.getAPI();
        }
        getLogger().log(Level.INFO, "Setting Up Vault now....");
        boolean canLoad = SetupVault();
        if (!canLoad) {
            getLogger().log(Level.INFO, "Vault Failed....");
            getServer().getPluginManager().disablePlugin(this);
            return;
        }
        CitizensAPI.getTraitFactory().registerTrait(net.citizensnpcs.api.trait.TraitInfo.create(BlacksmithTrait.class).withName("blacksmith"));

        getLogger().log(Level.INFO, " v" + getDescription().getVersion() + " enabled.");
    }

    private boolean SetupVault() {
        // Setup Vault
        RegisteredServiceProvider<Economy> economyProvider = getServer().getServicesManager().getRegistration(
                Economy.class);
        if (economyProvider != null) {
            economy = economyProvider.getProvider();
            return true;
        } else {
            // Disable if no economy plugin was found
            getServer().getLogger().log(Level.SEVERE, "Failed to load an economy plugin. Disabling...");
            return false;
        }
    }

    @Override
    public boolean onCommand(final CommandSender sender, final @NotNull Command command, final @NotNull String label, final String[] args) {
        config.load();
        sender.sendMessage(ChatColor.GREEN + "Blacksmith config reloaded!");
        return true;
    }

    public boolean isTool(ItemStack item) {
        return switch (item.getType()) {
            case WOODEN_PICKAXE, WOODEN_SHOVEL, WOODEN_HOE, WOODEN_SWORD, WOODEN_AXE, STONE_PICKAXE, STONE_SHOVEL,
                    STONE_HOE, STONE_SWORD, STONE_AXE, GOLDEN_PICKAXE, GOLDEN_SHOVEL, GOLDEN_HOE, GOLDEN_SWORD,
                    GOLDEN_AXE, IRON_PICKAXE, IRON_SHOVEL, IRON_HOE, IRON_SWORD, IRON_AXE, DIAMOND_PICKAXE,
                    DIAMOND_SHOVEL, DIAMOND_HOE, DIAMOND_SWORD, DIAMOND_AXE, NETHERITE_SWORD, NETHERITE_SHOVEL,
                    NETHERITE_PICKAXE, NETHERITE_AXE, NETHERITE_HOE, BOW, CROSSBOW, FLINT_AND_STEEL, FISHING_ROD,
                    SHEARS, TRIDENT, SHIELD -> true;
            default -> false;
        };
    }

    public boolean isArmor(ItemStack item) {
        return switch (item.getType()) {
            case LEATHER_HELMET, LEATHER_CHESTPLATE, LEATHER_LEGGINGS, LEATHER_BOOTS, CHAINMAIL_HELMET,
                    CHAINMAIL_CHESTPLATE, CHAINMAIL_LEGGINGS, CHAINMAIL_BOOTS, GOLDEN_HELMET, GOLDEN_CHESTPLATE,
                    GOLDEN_LEGGINGS, GOLDEN_BOOTS, IRON_HELMET, IRON_CHESTPLATE, IRON_LEGGINGS, IRON_BOOTS,
                    DIAMOND_HELMET, DIAMOND_CHESTPLATE, DIAMOND_LEGGINGS, DIAMOND_BOOTS, TURTLE_HELMET, ELYTRA,
                    NETHERITE_HELMET, NETHERITE_CHESTPLATE, NETHERITE_LEGGINGS, NETHERITE_BOOTS -> true;
            default -> false;
        };
    }

    public boolean doesPlayerHaveEnough(Player player) {
        return economy.getBalance(player) - getCost(player.getInventory().getItemInMainHand(), player) >= 0;
    }

    public String formatCost(Player player) {
        double cost = getCost(player.getInventory().getItemInMainHand(), player);
        return economy.format(cost);
    }

    public void withdraw(Player player) {
        economy.withdrawPlayer(player, getCost(player.getInventory().getItemInMainHand(), player));
    }

    private double getCost(ItemStack item, Player player) {
        DataKey root = config.getConfig().getKey("");
        double price = Setting.BASE_PRICE.asDouble();
        if (root.keyExists("base-prices." + item.getType().name().toLowerCase().replace('_', '-'))) {
            price = root.getDouble("base-prices." + item.getType().name().toLowerCase().replace('_', '-'));
        }

        // Adjust price based on durability and enchantments
        if (this.useHyperAPI) {
            // If using hyperconomy, price is calculated like so:
            // New Item Price + Enchantments Price (from hyperconomy) / maxDurability = price per durability point
            // Total price would then be base_price + price per durablity point * current durability
            double hyperPrice = 0;
            HItemStack hi = hyperAPI.getHyperPlayer(player.getName()).getItemInHand();
            ItemStack item2 = player.getInventory().getItemInMainHand().clone();

            for (TradeObject enchant : hyperAPI.getEnchantmentHyperObjects(hi, player.getName())) {
                hyperPrice = hyperPrice + enchant.getBuyPrice(1);
                Enchantment enchantment = Enchantment.getByKey(NamespacedKey.minecraft(enchant.getEnchantment().getEnchantmentName()));
                item2.removeEnchantment(Objects.requireNonNull(enchantment));
            }

            ArrayList<Material> leathers = new ArrayList<>();
            leathers.add(Material.LEATHER_BOOTS);
            leathers.add(Material.LEATHER_CHESTPLATE);
            leathers.add(Material.LEATHER_HELMET);
            leathers.add(Material.LEATHER_LEGGINGS);

            HItemStack hi3 = null;
            if (leathers.contains(player.getInventory().getItemInMainHand().getType())) {
                hi3 = bukCon.getBukkitCommon().getSerializableItemStack(new ItemStack(player.getInventory().getItemInMainHand().getType()));
            }

            TradeObject to = this.hyperAPI.getHyperObject(hi, "default");
            if (to == null) {
                to = hyperAPI.getHyperObject(hi3, "default");
                if (to == null) {
                    HItemStack hi4 = bukCon.getBukkitCommon().getSerializableItemStack(new ItemStack(player.getInventory().getItemInMainHand().getType()));
                    to = this.hyperAPI.getHyperObject(hi4, "default");
                }
                hyperPrice = hyperPrice + to.getSellPrice(1);

            } else {
                hyperPrice = to.getSellPrice(1);
            }
            double hyperPricePerDurability = hyperPrice / item.getType().getMaxDurability();
            price += (((Damageable) item).getDamage() * hyperPricePerDurability);

            double enchantmentModifier = Setting.ENCHANTMENT_MODIFIER.asDouble();
            for (Enchantment enchantment : item2.getEnchantments().keySet()) {
                if (root.keyExists("enchantment-modifiers." + enchantment.getKey().asString().toLowerCase().replace('_', '-'))) {
                    enchantmentModifier = root.getDouble("enchantment-modifiers."
                            + enchantment.getKey().asString().toLowerCase().replace('_', '-'));
                }
                price += enchantmentModifier * item2.getEnchantmentLevel(enchantment);
            }


            return price;
        } else {
            if (root.keyExists("price-per-durability-point." + item.getType().name().toLowerCase().replace('_', '-'))) {
                price += ((Damageable) item).getDamage() * root.getDouble("price-per-durability-point." + item.getType().name().toLowerCase().replace('_', '-'));
            } else {
                price += (((Damageable) item).getDamage() * Setting.PRICE_PER_DURABILITY_POINT.asDouble());
            }
        }

        double enchantmentModifier = Setting.ENCHANTMENT_MODIFIER.asDouble();
        for (Enchantment enchantment : item.getEnchantments().keySet()) {
            if (root.keyExists("enchantment-modifiers." + enchantment.getKey().asString().toLowerCase().replace('_', '-'))) {
                enchantmentModifier = root.getDouble("enchantment-modifiers."
                        + enchantment.getKey().asString().toLowerCase().replace('_', '-'));
            }
            price += enchantmentModifier * item.getEnchantmentLevel(enchantment);
        }
        return price;
    }
}
